@@ -1,39 +1,23 @@
-# Copyright (c) 2012-2015 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
-# Copyright (c) 2013-2014 Google, Inc.
-# Copyright (c) 2014-2020 Claudiu Popa <pcmanticore@gmail.com>
-# Copyright (c) 2014 Eevee (Alex Munroe) <amunroe@yelp.com>
-# Copyright (c) 2015-2016 Ceridwen <ceridwenv@gmail.com>
-# Copyright (c) 2015 Dmitry Pribysh <dmand@yandex.ru>
-# Copyright (c) 2015 David Shea <dshea@redhat.com>
-# Copyright (c) 2015 Philip Lorenz <philip@bithub.de>
-# Copyright (c) 2016 Jakub Wilk <jwilk@jwilk.net>
-# Copyright (c) 2016 Mateusz Bysiek <mb@mbdev.pl>
-# Copyright (c) 2017 Hugo <hugovk@users.noreply.github.com>
-# Copyright (c) 2017 Łukasz Rogalski <rogalski.91@gmail.com>
-# Copyright (c) 2018 Ville Skyttä <ville.skytta@iki.fi>
-# Copyright (c) 2019 Ashley Whetter <ashley@awhetter.co.uk>
-# Copyright (c) 2020 hippo91 <guillaume.peillex@gmail.com>
-# Copyright (c) 2020 Ram Rachum <ram@rachum.com>
-# Copyright (c) 2021 Pierre Sassoulas <pierre.sassoulas@gmail.com>
-# Copyright (c) 2021 Andrew Haigh <hello@nelf.in>
-# Copyright (c) 2021 pre-commit-ci[bot] <bot@noreply.github.com>
-# Copyright (c) 2021 Marc Mueller <30130371+cdce8p@users.noreply.github.com>
-
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
-# For details: https://github.com/PyCQA/astroid/blob/master/LICENSE
+# For details: https://github.com/PyCQA/astroid/blob/main/LICENSE
+# Copyright (c) https://github.com/PyCQA/astroid/blob/main/CONTRIBUTORS.txt
 
 """Astroid hooks for the Python standard library."""
 
 import functools
 import keyword
 from textwrap import dedent
+from typing import Iterator, List, Optional, Tuple
 
+import astroid
 from astroid import arguments, inference_tip, nodes, util
 from astroid.builder import AstroidBuilder, extract_node
+from astroid.context import InferenceContext
 from astroid.exceptions import (
     AstroidTypeError,
     AstroidValueError,
     InferenceError,
+    MroError,
     UseInferenceDefault,
 )
 from astroid.manager import AstroidManager
@@ -54,11 +38,11 @@ def _infer_first(node, context):
         raise UseInferenceDefault
     try:
         value = next(node.infer(context=context))
-        if value is util.Uninferable:
-            raise UseInferenceDefault()
-        return value
     except StopIteration as exc:
         raise InferenceError from exc
+    if value is util.Uninferable:
+        raise UseInferenceDefault()
+    return value
 
 
 def _find_func_form_arguments(node, context):
@@ -84,7 +68,12 @@ def _find_func_form_arguments(node, context):
     raise UseInferenceDefault()
 
 
-def infer_func_form(node, base_type, context=None, enum=False):
+def infer_func_form(
+    node: nodes.Call,
+    base_type: nodes.NodeNG,
+    context: Optional[InferenceContext] = None,
+    enum: bool = False,
+) -> Tuple[nodes.ClassDef, str, List[str]]:
     """Specific inference function for namedtuple or Python 3 enum."""
     # node is a Call node, class name as first argument and generated class
     # attributes as second argument
@@ -96,10 +85,13 @@ def infer_func_form(node, base_type, context=None, enum=False):
         try:
             attributes = names.value.replace(",", " ").split()
         except AttributeError as exc:
+            # Handle attributes of NamedTuples
             if not enum:
                 attributes = [
                     _infer_first(const, context).value for const in names.elts
                 ]
+
+            # Handle attributes of Enums
             else:
                 # Enums supports either iterator of (name, value) pairs
                 # or mappings.
@@ -141,10 +133,13 @@ def infer_func_form(node, base_type, context=None, enum=False):
     # we know it is a namedtuple anyway.
     name = name or "Uninferable"
     # we want to return a Class node instance with proper attributes set
-    class_node = nodes.ClassDef(name, "docstring")
-    class_node.parent = node.parent
-    # set base class=tuple
-    class_node.bases.append(base_type)
+    class_node = nodes.ClassDef(name, parent=node.parent)
+    class_node.postinit(
+        # set base class=tuple
+        bases=[base_type],
+        body=[],
+        decorators=None,
+    )
     # XXX add __init__(*attributes) method
     for attr in attributes:
         fake_node = nodes.EmptyNode()
@@ -177,17 +172,24 @@ _looks_like_enum = functools.partial(_looks_like, name="Enum")
 _looks_like_typing_namedtuple = functools.partial(_looks_like, name="NamedTuple")
 
 
-def infer_named_tuple(node, context=None):
+def infer_named_tuple(
+    node: nodes.Call, context: Optional[InferenceContext] = None
+) -> Iterator[nodes.ClassDef]:
     """Specific inference function for namedtuple Call node"""
     tuple_base_name = nodes.Name(name="tuple", parent=node.root())
     class_node, name, attributes = infer_func_form(
         node, tuple_base_name, context=context
     )
     call_site = arguments.CallSite.from_call(node, context=context)
-    func = next(extract_node("import collections; collections.namedtuple").infer())
+    node = extract_node("import collections; collections.namedtuple")
+    try:
+
+        func = next(node.infer())
+    except StopIteration as e:
+        raise InferenceError(node=node) from e
     try:
         rename = next(call_site.infer_argument(func, "rename", context)).bool_value()
-    except InferenceError:
+    except (InferenceError, StopIteration):
         rename = False
 
     try:
@@ -207,27 +209,21 @@ def infer_named_tuple(node, context=None):
         for index, name in enumerate(attributes)
     )
     fake = AstroidBuilder(AstroidManager()).string_build(
-        """
-class %(name)s(tuple):
+        f"""
+class {name}(tuple):
     __slots__ = ()
-    _fields = %(fields)r
+    _fields = {attributes!r}
     def _asdict(self):
         return self.__dict__
     @classmethod
     def _make(cls, iterable, new=tuple.__new__, len=len):
         return new(cls, iterable)
-    def _replace(self, %(replace_args)s):
+    def _replace(self, {replace_args}):
         return self
     def __getnewargs__(self):
         return tuple(self)
-%(field_defs)s
+{field_defs}
     """
-        % {
-            "name": name,
-            "fields": attributes,
-            "field_defs": field_defs,
-            "replace_args": replace_args,
-        }
     )
     class_node.locals["_asdict"] = fake.body[0].locals["_asdict"]
     class_node.locals["_make"] = fake.body[0].locals["_make"]
@@ -349,9 +345,7 @@ INT_FLAG_ADDITION_METHODS = """
 
 def infer_enum_class(node):
     """Specific inference for enums."""
-    for basename in node.basenames:
-        # TODO: doesn't handle subclasses yet. This implementation
-        # is a hack to support enums.
+    for basename in (b for cls in node.mro() for b in cls.basenames):
         if basename not in ENUM_BASE_NAMES:
             continue
         if node.root().name == "enum":
@@ -363,7 +357,7 @@ def infer_enum_class(node):
             if any(not isinstance(value, nodes.AssignName) for value in values):
                 continue
 
-            stmt = values[0].statement()
+            stmt = values[0].statement(future=True)
             if isinstance(stmt, nodes.Assign):
                 if isinstance(stmt.targets[0], nodes.Tuple):
                     targets = stmt.targets[0].itered()
@@ -412,9 +406,9 @@ def infer_enum_class(node):
                     # should result in some nice symbolic execution
                     classdef += INT_FLAG_ADDITION_METHODS.format(name=target.name)
 
-                fake = AstroidBuilder(AstroidManager()).string_build(classdef)[
-                    target.name
-                ]
+                fake = AstroidBuilder(
+                    AstroidManager(), apply_transforms=False
+                ).string_build(classdef)[target.name]
                 fake.parent = target.parent
                 for method in node.mymethods():
                     fake.locals[method.name] = [method]
@@ -471,17 +465,20 @@ def infer_typing_namedtuple_class(class_node, context=None):
     """
     ).format(typename=class_node.name, fields=",".join(annassigns_fields))
     node = extract_node(code)
-    generated_class_node = next(infer_named_tuple(node, context))
+    try:
+        generated_class_node = next(infer_named_tuple(node, context))
+    except StopIteration as e:
+        raise InferenceError(node=node, context=context) from e
     for method in class_node.mymethods():
         generated_class_node.locals[method.name] = [method]
 
-    for assign in class_node.body:
-        if not isinstance(assign, nodes.Assign):
-            continue
-
-        for target in assign.targets:
-            attr = target.name
-            generated_class_node.locals[attr] = class_node.locals[attr]
+    for body_node in class_node.body:
+        if isinstance(body_node, nodes.Assign):
+            for target in body_node.targets:
+                attr = target.name
+                generated_class_node.locals[attr] = class_node.locals[attr]
+        elif isinstance(body_node, nodes.ClassDef):
+            generated_class_node.locals[body_node.name] = [body_node]
 
     return iter((generated_class_node,))
 
@@ -501,13 +498,15 @@ def infer_typing_namedtuple_function(node, context=None):
     return klass.infer(context)
 
 
-def infer_typing_namedtuple(node, context=None):
+def infer_typing_namedtuple(
+    node: nodes.Call, context: Optional[InferenceContext] = None
+) -> Iterator[nodes.ClassDef]:
     """Infer a typing.NamedTuple(...) call."""
     # This is essentially a namedtuple with different arguments
     # so we extract the args and infer a named tuple.
     try:
         func = next(node.func.infer())
-    except InferenceError as exc:
+    except (InferenceError, StopIteration) as exc:
         raise UseInferenceDefault from exc
 
     if func.qname() != "typing.NamedTuple":
@@ -529,13 +528,25 @@ def infer_typing_namedtuple(node, context=None):
 
     typename = node.args[0].as_string()
     if names:
-        field_names = "({},)".format(",".join(names))
+        field_names = f"({','.join(names)},)"
     else:
         field_names = "''"
     node = extract_node(f"namedtuple({typename}, {field_names})")
     return infer_named_tuple(node, context)
 
 
+def _is_enum_subclass(cls: astroid.ClassDef) -> bool:
+    """Return whether cls is a subclass of an Enum."""
+    try:
+        return any(
+            klass.name in ENUM_BASE_NAMES
+            and getattr(klass.root(), "name", None) == "enum"
+            for klass in cls.mro()
+        )
+    except MroError:
+        return False
+
+
 AstroidManager().register_transform(
     nodes.Call, inference_tip(infer_named_tuple), _looks_like_namedtuple
 )
@@ -543,11 +554,7 @@ AstroidManager().register_transform(
     nodes.Call, inference_tip(infer_enum), _looks_like_enum
 )
 AstroidManager().register_transform(
-    nodes.ClassDef,
-    infer_enum_class,
-    predicate=lambda cls: any(
-        basename for basename in cls.basenames if basename in ENUM_BASE_NAMES
-    ),
+    nodes.ClassDef, infer_enum_class, predicate=_is_enum_subclass
 )
 AstroidManager().register_transform(
     nodes.ClassDef, inference_tip(infer_typing_namedtuple_class), _has_namedtuple_base
